# -*- coding: utf-8 -*-
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.naive_bayes import MultinomialNB
from sklearn.metrics import confusion_matrix
from sklearn.metrics import accuracy_score
import numpy as np
import sys
import io
import string
from math import log
import time
import pickle

class mainMethod():
    def __init__(self):
        #used to time the duration of program execution
        start = time.time()

        global trainSize
        global devSize
        global testDataSize
        global createTestCSV

        # Number of data used for training
        trainCount = 250000

        #check command line arguments
        mode = '-mode1'         #default: mode1
        createTestCSV = False   #default: no export
        #These are only important for mode2:

        m2method = '-demo'      #default: run decisiontree in demo version


        for arg in sys.argv:
            if (arg == '-mode1' or arg == '-mode2' or arg =='-mode3'):
                mode = arg
            elif (arg == '-export'):
                createTestCSV = True
            elif (arg == '-demo' or arg == '-trainFull' or arg =='-load'):
                m2method = arg


        #set random seed arbitrarily to secure same performance on every run
        np.random.seed(135918274)
        #save the random state to ensure data and classification get shuffled the same.
        randState = np.random.get_state()

        #initialize lists
        data = []
        y = []
        testData = []


        #read the data
        self.openx(data)
        self.openy(y)
        self.opentest(testData)

        np.random.shuffle(data)
        #resetting random state to get same shuffle
        np.random.set_state(randState)
        np.random.shuffle(y)


        #splitting up data into training and development set:

        train = data[0:trainCount]
        yTrain = y[0:trainCount]
        dev = data[trainCount:]
        yDev = y[trainCount:]

        trainSize = len(train)
        devSize = len(dev)
        testDataSize = len(testData)
        print("Sentences used for Training: ", trainSize)
        print("Sentences used for Development: ", devSize)
        print("Sentences in Test set: ", testDataSize)

        #get the feature representation of the sentences
        self.doVectorize(train, dev, testData)

        if (mode == '-mode1'):
            self.doPart1(train,testData,yTrain,yDev)
        if (mode == '-mode2'):
            self.doPart2(m2method,train,dev,testData,yTrain,yDev)
        if (mode == '-mode3'):
            self.doPart3(yTrain, yDev)

        #self.checkOutput(y,'outputPart1.csv')
        print ("Time needed: {}s".format(time.time()-start))


    def openx(self, data):
        #opens the train set csv file and reads the important information.
        with io.open('train_set_x.csv', 'r', encoding='utf-8') as f:
            # get rid of first line
            f.readline()
            raw = f.readlines()
            for line in raw:
                splitline = line.split(',')
                #remove end of line characters
                splitline[1] = str.replace(splitline[1], '\n', '')
                # remove space characters
                splitline[1] = str.replace(splitline[1], ' ', '')
                # only get text and not id
                data += [splitline[1]]


    def openy(self, y):
        #opens the train set classification csv file and reads the important information.
        with io.open('train_set_y.csv', 'r', encoding='utf-8') as f:
            # get rid of first line
            f.readline()
            raw = f.readlines()

            for i, line in enumerate(raw):
                splitline = line.split(',')
                splitline[1] = str.replace(splitline[1], '\n', '')
                # convert to int for convinience
                y += [int(splitline[1])]

    def opentest(self,testData):
        #imports the test_set data
        with io.open('test_set_x.csv', 'r', encoding='utf-8') as f:
            # get rid of first line
            f.readline()
            raw = f.readlines()
            for line in raw:
                splitline = line.split(',')
                splitline[1] = str.replace(splitline[1], '\n', '')
                # remove space characters
                splitline[1] = str.replace(splitline[1], ' ', '')
                # only get text and not id
                testData += [splitline[1]]

    def doVectorize(self, train,dev, test):
        global vectorizer
        #CountVectorizer used to convert the character strings into a feature vector.
        #creates a feature for every different unicode character observed.

        vectorizer = CountVectorizer(analyzer="char")
        global trainFeatures
        global devFeatures
        global testFeatures

        #Feature Vectors for training and development and test sentences
        trainFeatures = vectorizer.fit_transform(train)
        devFeatures = vectorizer.transform(dev)
        testFeatures = vectorizer.transform(test)


    def doPart1(self,train,test,yTrain,yDev):
        #Implementation of Naive Bayes
        print ("\nUsing Manually Implemented Naive Bayes: \n")
        #Training:
        #Finding log (P(y)):

        #count how often each class occurs
        classCount = [0,0,0,0,0]
        for idx ,sample in enumerate(train):
            classCount[int(yTrain[idx])] += 1
        #Calculating log(P(y))
        classProbs = [0.0,0.0,0.0,0.0,0.0]
        for idx,clas in enumerate(classCount):
            #devide classCounts by number of samples and log
            classProbs[idx] = log(clas/(trainSize * 1.0))

        #Finding P(X|y):

        features = vectorizer.get_feature_names()
        # includes lapace smoothing. Used to add sum of different characters to total char counts.
        l = len(features)
        totals = [l,l,l,l,l]
        #get total amount of letters per class

        #convert to array to allow numpy magic
        y2 = np.array(yTrain)


        bigLetterCount = []
        for clas in range(0,5):
            # bigLetterCount saves the Counts of all the letters per class.
            # Syntax: bigLetterCount[class][idx of letter]
            charOfClass = np.squeeze(np.asarray(np.sum(trainFeatures[y2 == clas], axis=0)))
            totals[clas] += np.sum(charOfClass)
            bigLetterCount.append(charOfClass)


        letterProbsAllClasses = []
        #get relative frequencys of characters and log
        for clas,letters in enumerate(bigLetterCount):
            #laplace smoothing adds 1 to each occurence to avoid -Inf prob
            #devide occurance of character in each class by total characters in each class.
            letterProbsOneClass = np.log((letters +1.0) / np.array(totals[clas]))
            letterProbsAllClasses += [letterProbsOneClass]

        #log prob matrix created.

        
        #prediction of train set (to calculate training error):

        logs = []
        for clas in range(0,5):
            #Calculate the dotproduct of the feature matrix and the probability vector
            #Corresponds to a sum of all the log probs of characters that occur
            logs.append(trainFeatures.dot(letterProbsAllClasses[clas]) + classProbs[clas])

        #convert to matrix to allow numpy magic
        logs = np.asmatrix(logs)
        #get the class with the highest log likelihood
        classification = np.squeeze(np.asarray(np.argmax(logs,axis=0)))

        print("Training Accuracy:")
        print(accuracy_score(yTrain,classification))



        # prediction of development set (validation):
        logs = []
        for clas in range(0, 5):
            # Calculate the dotproduct of the feature matrix and the probability vector
            # Corresponds to a sum of all the log probs of characters that occur
            logs.append(devFeatures.dot(letterProbsAllClasses[clas]) + classProbs[clas])

        # convert to matrix to allow numpy magic
        logs = np.asmatrix(logs)
        # get the class with the highest log likelihood
        classification = np.squeeze(np.asarray(np.argmax(logs, axis=0)))
        print("Validation Accuracy:")
        print(accuracy_score(yDev, classification))
        print("Validation Confusion Matrix:")
        print(confusion_matrix(yDev, classification))

        #if output for testset is wanted
        if (createTestCSV):
            # prediction of test set:
            logs = []
            for clas in range(0, 5):
                # Calculate the dotproduct of the feature matrix and the probability vector
                # Corresponds to a sum of all the log probs of characters that occur
                logs.append(testFeatures.dot(letterProbsAllClasses[clas]) + classProbs[clas])
            # convert to matrix to allow numpy magic
            logs = np.asmatrix(logs)
            # get the class with the highest log likelihood
            classification = np.squeeze(np.asarray(np.argmax(logs, axis=0)))
            #output the classification of test set
            output = []
            for i,clas in enumerate(classification):
                output += u"{},{}\n".format(i, clas)
            ###save output###
            with io.open('outputMode1.csv', 'w', encoding='utf-8') as f:
                #create header
                f.write(u"Id,Category\n")
                for o in output:
                    f.write(o)

    def doPart2(self,m2method, train,dev,test,yTrain,yDev):
        print ("\n\nUsing manually implemented decision tree: \n")

        #reduce training samples in demonstration case.
        if (m2method== '-demo'):
            train = train[:1000]
            yTrain= yTrain[:1000]


        # helper function that checks if a character is a number
        def is_number(s):
            try:
                float(s)
                return True
            except ValueError:
                pass
            return False

        #For the decision tree, we use sets of characters occuring in the sentences and ignore counts
        #get sets of used characters in training set

        allTrain = []
        for idx,i in enumerate(train):
            temp = []
            # get rid of space and number from training data
            if len(i) != 0:
                uniqueSet = set(list(i))
                #remove numbers
                temp = [item for item in uniqueSet if is_number(item) == False]

            temp.append(str(yTrain[idx]))
            allTrain.append(temp)

        #get sets of used characters in development set
        allDev = []
        for idx,i in enumerate(dev):
            temp = []
            if len(i) != 0:
                uniqueSet = set(list(i))
                # remove numbers
                temp = [item for item in uniqueSet if is_number(item) == False]

            temp.append(str(yDev[idx]))
            allDev.append(temp)

        #gets sets of used characters in test set
        allTest = []
        for i in test:
            temp = []
            # get rid of space and number from test data
            if len(i) != 0:
                uniqueSet = set([letter for letter in i if letter != ' '])
                # remove numbers
                temp = [item for item in uniqueSet if is_number(item) == False]

            allTest.append(temp)


        ##return the count of each class as a dictionary
        def classCount(allTrain):
            numClass = {}
            for item in allTrain:
                if item[-1] not in numClass:
                    numClass[item[-1]] = 0

                numClass[item[-1]] += 1

            return numClass


            ##split training data based on presence/absence of feature in each sample

        def divide(allTrain, letter, featValue):
            set1 = [sentence for sentence in allTrain if (letter in sentence) == featValue]
            set2 = [sentence for sentence in allTrain if (letter in sentence) != featValue]
            return set1, set2

        ##calculate entropy of training data
        def entropy(allTrain):
            nSentences = len(allTrain)
            nClass = classCount(allTrain)
            ent = 0.0
            for group in nClass.keys():
                p = float(nClass[group]) / nSentences
                if p != 0:
                    ent -= p * (log(p) / log(2))
            return ent

        ##extract all the features in training data except numbers
        def extractFeatures(allTrain):
            letters = []
            for sentence in allTrain:
                for letter in sentence[:-1]:
                    if is_number(letter) == False and letter not in letters:
                        letters.append(letter)
            return letters

        ##assign label of major class to leaf based on result of majority vote
        def majorityCount(classLabel):
            count = {}
            for label in classLabel:
                if label not in count:
                    count[label] = 0
                count[label] += 1
                sortedCount = sorted(count.items(), key=lambda x: x[1], reverse=True)
            return sortedCount[0][0]

        ##top-down recursively build decision tree
        def buildTree(allTrain, features):
            bestGain = 0.0
            bestFeature = None
            bestSets = None

            classLabel = [item[-1] for item in allTrain]

            # stop splitting when there is only one sample in training data
            if len(allTrain) == 1:
                return majorityCount(classLabel)

            # stop splitting when all samples in training data are the same classes
            if classLabel.count(classLabel[0]) == len(classLabel):
                return classLabel[0]

            # find the feature that gets the best information gain and split samples based on this best feature
            for letter in features:
                set1 = divide(allTrain, letter, 1)[0]
                set2 = divide(allTrain, letter, 1)[1]

                numSentence = len(allTrain)
                baseEntropy = entropy(allTrain)
                p = float(len(set1)) / numSentence

                entropy1 = entropy(set1)
                entropy2 = entropy(set2)

                gain = baseEntropy - p * entropy1 - (1 - p) * entropy2

                if gain > bestGain and len(set1) > 0 and len(set2) > 0:
                    bestGain = gain
                    bestFeature = letter
                    bestSets = (set1, set2)

            # initialize a tree dictionary
            tree = {bestFeature: {}}

            # exclude best feature from next recursion of data splitting
            if bestFeature in features:
                features.remove(bestFeature)

            # check how many different classes are inside a training data, this will decide how many keys to use in tree dictionary
            # 0: no sample in the data contain bestFeature
            # 1: all samples in the data contain bestFeature
            # 1 and 0: some samples inside the data contain bestFeature, some not
            booleanFeature = []
            count = 0
            for sentence in allTrain:
                if bestFeature in sentence:
                    count += 1
            if count == 0:
                booleanFeature = [0]
            elif count == len(train):
                booleanFeature = [1]
            else:
                booleanFeature = [0, 1]

            # recursively split training data and build tree dictionary
            for value in booleanFeature:
                subFeature = [feat for feat in features]
                if bestSets != None:
                    allTrain = bestSets[value - 1]
                    tree[bestFeature][value] = buildTree(allTrain, subFeature)

            return tree

        ##based on presence/absence of each node feature in testsample, recursively trace down a tree from root until reaching a leaf node and assign that leaf's class to the testsample
        def classifier(tree, testSample):
            firstNode = list(tree.keys())[0]
            subTree = tree[firstNode]
            #If the sentence is empty default french (because its to most occuring)
            if firstNode == None:
                return u'1'
            if firstNode in testSample:
                key = 1
            else:
                key = 0

            value = subTree[key]

            if isinstance(value, dict):
                label = classifier(value, testSample)
            else:
                label = value

            return label

        ##store output in harddrive
        def storeTree(tree, myFile):
            outputFile = open(myFile, 'wb')
            pickle.dump(tree, outputFile)
            outputFile.close()

        ##load output from harddrive
        def grabTree(myFile):
            openFile = open(myFile,'rb')
            return pickle.load(openFile)


            ##calculate validation dataset classification error rate

        def errorCheck(myFile):
            errorCount = 0
            result = grabTree(myFile)

            for item in result:
                if item[0] != item[1]:
                    errorCount += 1

            return float(errorCount) / len(result)


        if (m2method == '-demo' or m2method == '-trainFull'):
            features = extractFeatures(allTrain)
            #This might take a while:
            tree = buildTree(allTrain, features)
            print("Tree built successfully")
            if (m2method == '-trainFull'):
                myFile = 'savedTree'
                storeTree(tree, myFile)

        if (m2method == '-load'):
            myFile = 'outputTreeDict_oct22.txt'
            tree = grabTree(myFile)

        # prediction of train set (to calculate training error):
        classification = []
        for i in allTrain:
            temp = i[:-1]
            label = classifier(tree, temp)
            classification.append(int(label))

        print("Training Accuracy:")
        print(accuracy_score(yTrain, classification))

        # prediction of development set (validation):
        classification = []
        for i in allDev:
            temp = i[:-1]
            label = classifier(tree, temp)
            classification.append(int(label))

        print("Validation Accuracy:")
        print(accuracy_score(yDev, classification))
        print("Validation Confusion Matrix:")
        print(confusion_matrix(yDev, classification))



        if (createTestCSV):
            # prediction of test set:
            classification = []
            for i in allDev:
                label = classifier(tree, i)
                classification.append(int(label))

            # output the classification of test set
            output = []
            for i, clas in enumerate(classification):
                output += u"{},{}\n".format(i, clas)
            ###save output###
            with io.open('outputMode2.csv', 'w', encoding='utf-8') as f:
                # create header
                f.write(u"Id,Category\n")
                for o in output:
                    f.write(o)


    def doPart3(self, yTrain, yDev):
        print ("\n\nUsing Sklearn Naive Bayes: \n")

        #Naive Bayes with Lidstone Smoothing

        clf = MultinomialNB(alpha = 0.2)
        #Train the Classifier
        clf.fit(trainFeatures, yTrain)


        # prediction of train set (to calculate training error):
        classification = clf.predict(trainFeatures)
        print("Training Accuracy:")
        print(accuracy_score(yTrain, classification))

        # prediction of development set (validation):
        classification = clf.predict(devFeatures)
        print("Validation Accuracy:")
        print(accuracy_score(yDev, classification))
        print("Validation Confusion Matrix:")
        print(confusion_matrix(yDev, classification))

        # if output for testset is wanted
        if (createTestCSV):
            # prediction of test set:
            classification = clf.predict(testFeatures)
            # output the classification of test set
            output = []
            for i, clas in enumerate(classification):
                output += u"{},{}\n".format(i, clas)
            ###save output###
            with io.open('outputMode3.csv', 'w', encoding='utf-8') as f:
                # create header
                f.write(u"Id,Category\n")
                for o in output:
                    f.write(o)

if __name__ == '__main__':
    mainMethod()
