# -*- coding: utf-8 -*-
import io
import string
from math import log
import random
import csv

with io.open('train_set_x.csv', 'r', encoding ='utf-8') as f:    	
    f.readline()
    raw = f.readlines()
    data = []
    for line in raw:
        splitline = line.split(',')
        splitline[1] = string.replace(splitline[1],'\n','')
        splitline[1] = " ".join(splitline[1])
            
        #only get text and not id
        data += [splitline]


with io.open('train_set_y.csv', 'r', encoding ='utf-8') as f:		
    f.readline()
    raw = f.readlines()
    for i,line in enumerate(raw):            
        splitline = line.split(',')
        splitline[1] = string.replace(splitline[1],'\n','')
        data[i] += [splitline[1]]
        
        
with io.open('test_set_x.csv', 'r', encoding ='utf-8') as f:
	f.readline()
	raw = f.readlines()
	testdata = []
	for line in raw:
		splitline = line.split(',')
		splitline[1] = string.replace(splitline[1],'\n','')
		splitline[1] = string.replace(splitline[1],' ','')
		testdata += [splitline]

       
##check whether the symbol is a number
def is_number(s):
    try:
        float(s)
        return True
    except ValueError:
        pass 
    return False


##store each training sample as a list contains unique symbols and sample ID 
allTrain = []
for i in data[:276516]:
    temp = []
    #get rid of space and number from training data 
    if len(i[1]) != 0:          
        uniqueSet = set([letter for letter in i[1] if letter != ' '])
        temp = [item for item in uniqueSet if is_number(item) == False]

    if len(temp) != 0:
        temp.extend(i[-1])
        allTrain.append(temp)


##randomly shuffle data and set aside first 20000 samples for validation test
random.shuffle(allTrain)
validate = allTrain[:20000]
train = allTrain[20000:]


##return the count of each class as a dictionary
def classCount(train):
    numClass = {}
    for item in train: 
        if item[-1] not in numClass:
            numClass[item[-1]] = 0
        
        numClass[item[-1]] += 1
        
    return numClass 
   

##split training data based on presence/absence of feature in each sample
def divide(train, letter, featValue):
    set1 = [sentence for sentence in train if (letter in sentence) == featValue]
    set2 = [sentence for sentence in train if (letter in sentence) != featValue]
    return set1, set2


    
##calculate entropy of training data
def entropy(train):
    nSentences = len(train)
    nClass = classCount(train)
    ent = 0.0
    for group in nClass.keys():
        p = float(nClass[group]) / nSentences
        if p!= 0:
            ent -= p * (log(p) / log(2))
    return ent


##extract all the features in training data except numbers
def extractFeatures(train):
    letters = []
    for sentence in train:
        for letter in sentence[:-1]:            
            if is_number(letter) == False and letter not in letters:
                letters.append(letter)
    return letters
    


##assign label of major class to leaf based on result of majority vote
def majorityCount(classLabel):
    count = {}
    for label in classLabel:
        if label not in count:
            count[label] = 0
        count[label] += 1
        sortedCount = sorted(count.items(), key = lambda x : x[1], reverse = True)
    return sortedCount[0][0]



##top-down recursively build decision tree 
def buildTree(train, features = extractFeatures(train)):
    bestGain = 0.0
    bestFeature = None    
    bestSets = None
        
    classLabel = [item[-1] for item in train]
    
    #stop splitting when there is only one sample in training data
    if len(train) == 1: 
        return majorityCount(classLabel)

    #stop splitting when all samples in training data are the same classes 
    if classLabel.count(classLabel[0]) == len(classLabel):
        return classLabel[0]
     
    #find the feature that gets the best information gain and split samples based on this best feature     
    for letter in features:
        set1 = divide(train, letter, 1)[0] 
        set2 = divide(train, letter, 1)[1]
                
        numSentence = len(train)
        baseEntropy = entropy(train)
        p = float(len(set1)) / numSentence

        entropy1 = entropy(set1)
        entropy2 = entropy(set2)

        gain = baseEntropy - p * entropy1 - (1-p) * entropy2

        if gain > bestGain and len(set1) > 0 and len(set2) > 0:
            bestGain = gain
            bestFeature = letter
            bestSets = (set1, set2)
        
    #initialize a tree dictionary
    tree = {bestFeature: {}}
    
    #exclude best feature from next recursion of data splitting
    if bestFeature in features:
        features.remove(bestFeature)
    
    #check how many different classes are inside a training data, this will decide how many keys to use in tree dictionary
    #0: no sample in the data contain bestFeature
    #1: all samples in the data contain bestFeature 
    #1 and 0: some samples inside the data contain bestFeature, some not     
    booleanFeature = []
    count = 0
    for sentence in train:
        if bestFeature in sentence:
            count += 1
    if count == 0:
        booleanFeature = [0]
    elif count == len(train):
        booleanFeature = [1]
    else:
        booleanFeature = [0, 1]
    
    #recursively split training data and build tree dictionary
    for value in booleanFeature:
        subFeature = [feat for feat in features] 
        if bestSets != None:
            train = bestSets[value-1]
            tree[bestFeature][value] = buildTree(train, subFeature)

    return tree


##based on presence/absence of each node feature in testsample, recursively trace down a tree from root until reaching a leaf node and assign that leaf's class to the testsample       
def classifier(tree, testSample):
    firstNode = tree.keys()[0]
    subTree = tree[firstNode]
    if firstNode == None:
        return u'1'
    if firstNode in testSample:
        key = 1
    else:
        key = 0 

    value = subTree[key]
    
    if isinstance(value, dict):
        label = classifier(value, testSample)
    else:
        label = value
        
    return label



##store output in harddrive
def storeTree(tree, myFile):
    import pickle
    outputFile = open(myFile, 'w')
    pickle.dump(tree, outputFile)
    outputFile.close()


    
##load output from harddrive   
def grabTree(myFile):
    import pickle
    openFile = open(myFile)
    return pickle.load(openFile)    


##calculate validation dataset classification error rate
def errorCheck(myFile):
    errorCount = 0
    result = grabTree(myFile)

    for item in result:
        if item[0] != item[1]:
            errorCount += 1

    return float(errorCount) / len(result)



#features = extractFeatures(train)
#tree = buildTree(train, features)
#
#myFile = 'outputTreeDict_oct22.txt'
#storeTree(tree, myFile)
#tree = grabTree(myFile)


##validation test: predict class label of each test sample and save (actual class, predicted class) as tuple in a list output
classOutput = []    
for i in validate:
    temp= i[:-1]
    label = classifier(tree, temp)
    classOutput.append((i[-1],label))
   

#targetFile = 'classifyOutput_oct22.txt'
#storeTree(classOutput, targetFile)
#outputFile = grabTree(targetFile)
#print errorCheck(targetFile)



##kaggle prediction: classify each testsample and save (sample ID, predicted class) as tuple in a list output
#testOutput = []    
#for i in testdata[:118508]:
#    temp = []
#    tempSet = set([item for item in i[1]])
#    temp= [letter for letter in tempSet]
#    label = classifier(tree, temp)
#    testOutput.append((i[0],label))
#
#
#testOutputFile = 'testPredictOutput_oct22.txt'
#storeTree(testOutput, testOutputFile)
#kaggleFile = grabTree(testOutputFile)

##convert classification result into csv format for uploading to kaggle and comparing predeiction
#with io.open('myClassifyOutput.csv', 'wb') as f:    
#    csvOut = csv.writer(f)
#    csvOut.writerow(['Id', 'Category'])
#    for o in kaggleFile:
#        csvOut.writerow(o)




